from typing import List, Tuple

import pytest

from selection_types.input_coin import InputCoin
from selection_types.output_group import OutputGroup


@pytest.fixture
def generate_utxo_pool():
    def _generate_utxo_pool(
        amounts: List[int],
        short_term_fee_per_byte: int = 0,
        long_term_fee_per_byte: int = 0
    ) -> List[OutputGroup]:
        utxo_pool: List[OutputGroup] = []
        for amount in amounts:
            input_coin = InputCoin(
                tx_hash="",
                vout=0,
                value=int(amount),
                input_bytes=100
            )
            output_group = OutputGroup([input_coin])
            output_group.set_fee(short_term_fee_per_byte,
                                 long_term_fee_per_byte)
            utxo_pool.append(output_group)
        return utxo_pool
    return _generate_utxo_pool


@pytest.fixture
def make_hard_case(generate_utxo_pool):
    def _make_hard_case(utxo_count: int) -> Tuple[int, List[OutputGroup]]:
        target_value = 0
        utxo_amounts: List[int] = []

        for i in range(utxo_count):
            target_value += 1 << (utxo_count + i)
            utxo_amount_1 = 1 << (utxo_count + i)
            utxo_amount_2 = (1 << (utxo_count + i)) + (1 << (utxo_count-1-i))
            utxo_amounts.append(utxo_amount_1)
            utxo_amounts.append(utxo_amount_2)

        utxo_pool = generate_utxo_pool(utxo_amounts)
        return (target_value, utxo_pool)

    return _make_hard_case



from typing import List

import pytest

from selection_algorithms.branch_and_bound import select_coins_branch_and_bound
from selection_types.change_constants import CENT
from selection_types.coin_selection import CoinSelection
from selection_types.input_coin import InputCoin
from selection_types.output_group import OutputGroup
from tests.fixtures import generate_utxo_pool, make_hard_case

# todo: run tests with retries and repeats
# todo: test cases for non-zero npon-input fees

# How many times to run all the tests to have a chance to catch errors
# that only show up with particular random shuffles
RUN_TESTS = 100
# Some tests fail 1% of the time due to bad luck.
# We repeat those tests this many times and only complain if all iterations of the test fail
RANDOM_REPEATS = 5

DEFAULT_NOT_INPUT_FEES = 0


@pytest.mark.parametrize("target_amount", [1 * CENT, 2 * CENT, 3 * CENT, 4 * CENT])
def test_branch_and_bound_exact_match_single_coin(generate_utxo_pool, target_amount):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, target_amount, 0.5 * CENT)

    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert len(selection.outputs) == 1
    assert selection.effective_value == target_amount


def test_branch_and_bound_insufficient_funds(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, 11 * CENT, 0.5 * CENT)

    assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE
    assert len(selection.outputs) == 0
    assert selection.effective_value == 0


# Cost of change is greater than the difference between target value and utxo sum
def test_branch_and_bound_expensive_change(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, 0.9 * CENT, 0.5 * CENT)

    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert len(selection.outputs) == 1
    assert selection.effective_value == 1 * CENT


# Cost of change is less than the difference between target value and utxo sum
def test_branch_and_bound_cheap_change_failure(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, 0.9 * CENT, 0)
    assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE
    assert len(selection.outputs) == 0
    assert selection.effective_value == 0


def test_branch_and_bound_exact_match_multiple_coins(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
        5 * CENT
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, 10 * CENT, 0.5 * CENT)
    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert len(selection.outputs) == 3
    assert selection.effective_value == 10 * CENT

    selected_amounts = [
        output.effective_value for output in selection.outputs]
    selected_amounts.sort()

    assert selected_amounts[0] == 1 * CENT
    assert selected_amounts[1] == 4 * CENT
    assert selected_amounts[2] == 5 * CENT


def test_branch_and_bound_failure_no_match(generate_utxo_pool):
    utxo_pool = generate_utxo_pool([
        1 * CENT,
        2 * CENT,
        3 * CENT,
        4 * CENT,
        5 * CENT
    ])

    selection = select_coins_branch_and_bound(
        utxo_pool, 0.25 * CENT, 0.5 * CENT)
    assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE
    assert len(selection.outputs) == 0
    assert selection.effective_value == 0


def test_branch_and_bound_iteration_exhaustion(make_hard_case):
    target_value, utxo_pool = make_hard_case(17)
    selection = select_coins_branch_and_bound(
        utxo_pool, target_value, 0)
    assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE
    assert len(selection.outputs) == 0

    target_value, utxo_pool = make_hard_case(14)
    selection = select_coins_branch_and_bound(
        utxo_pool, target_value, 0)
    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert len(selection.outputs) > 0


def test_branch_and_bound_early_bailout_optimization(generate_utxo_pool):
    utxo_pool = generate_utxo_pool(
        [2 * CENT,
         7 * CENT,
         7 * CENT,
         7 * CENT,
         7 * CENT] +
        [5 * CENT for i in range(50000)]
    )
    selection = select_coins_branch_and_bound(
        utxo_pool, 30 * CENT, 5000)
    assert selection.outcome == CoinSelection.Outcome.SUCCESS
    assert len(selection.outputs) == 5
    assert selection.effective_value == 30 * CENT


def test_branch_and_bound_consistently_fails_impossible_case(generate_utxo_pool):
    utxo_pool = generate_utxo_pool(
        [i * CENT for i in range(5, 21)]
    )
    for i in range(100):
        selection = select_coins_branch_and_bound(
            utxo_pool, 1 * CENT, 2 * CENT)
        assert selection.outcome == CoinSelection.Outcome.ALGORITHM_FAILURE
        assert len(selection.outputs) == 0
        assert selection.effective_value == 0



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bitcoin_coin_selection",
    version="1.0.0",
    author="Kevin Gislason",
    author_email="kevingislason@gmail.com",
    description="Port of Bitcoin core coin selection logic to Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kevingislason/bitcoin_coin_selection",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries"
    ],
    python_requires='>=3.6',
)



from typing import List
from selection_types.input_coin import InputCoin

# See https://github.com/bitcoin/bitcoin/pull/12257
"""
Each output group corresponds to all outputs with the same destination
It is a privacy improvement, as each time you spend some output, any other
output that is publicly associated with the destination (address) will also be
spent at the same time, at the cost of fee increase for cases where coin
select without group restriction would find a more optimal set of coins
"""


class OutputGroup():
    outputs: List[InputCoin]
    value: int
    effective_value: int
    fee: int
    long_term_fee: int

    def __init__(self, outputs=None):
        self.effective_value = None
        self.fee = None
        self.long_term_fee = None
        self.value = 0
        self.outputs = []
        if outputs:
            for output in outputs:
                self.insert(output)

    def __gt__(self, output_group_2):
        return self.effective_value > output_group_2.effective_value

    def set_fee(self, short_term_fee_per_byte, long_term_fee_per_byte):
        self.effective_value = 0
        self.fee = 0
        self.long_term_fee = 0

        non_negative_ev_outputs = []
        for output in self.outputs:
            output.set_fee(short_term_fee_per_byte, long_term_fee_per_byte)
            # Filter outputs with negative effective values
            if output.effective_value > 0:
                self.fee += output.fee
                self.long_term_fee += output.long_term_fee
                self.effective_value += output.effective_value
                non_negative_ev_outputs.append(output)

    def insert(self, output: InputCoin):
        self.outputs.append(output)
        self.value += output.value
